#include <ArduinoJson.h>
#include <WebServer.h>

#include "ESPTelnet.h"

#include <TaskScheduler.h>

#include <Adafruit_Sensor.h>
#include <Adafruit_BME280.h>

// --------- WIFI -----------
#define STASSID    "WIFI-NAME"
#define STAPSK     "WIFI-PWD"

#define DEVICENAME "ESP32-DataLogger";

unsigned long previousMillis = 0;
unsigned long interval = 2000;

// --------- END WIFI -------

// --------- INITS -------

const char* ssid = STASSID;
const char* password = STAPSK;
const char* deviceName = DEVICENAME;
WebServer server(80);
StaticJsonDocument<1024> jsonDocument;
char jsonBuffer[1024];

ESPTelnet telnet;

// --------- END INITS -------

// --------- SCHEDULER BEGIN -------

void checkFreeRam();
Task scheduleCheckFreeRam(21*1000, TASK_FOREVER, &checkFreeRam);

void wifiReconnectCheck();
Task scheduleWifiReconnectCheck(5*1000, TASK_FOREVER, &wifiReconnectCheck);

void getBme();
Task scheduleGetBme(50000, TASK_FOREVER, &getBme);

void getValues();
Task scheduleGetValues(300, TASK_FOREVER, &getValues);

Scheduler runner;

// --------- SCHEDULER END ---------

// ------- DEFINITIONS ----------
static String bme280State = "State";

static String bme280Enabled = "enabled";
static String bme280Disabled = "disabled";

static String temperatureName = "Temperature";
static String temperatureName2 = "&deg;C";

static String humidityName = "Humidity";
static String humidityName2 = "&percnt;";

static String pressureName = "Pressure";
static String pressureName2 = "hPa";

static String linkColorNormal = "#2321B0";
static String linkColorVisited = "#2321B0";
static String activeMarkerBegin = "<b>&raquo;";
static String activeMarkerEnd = "&laquo;</b>";

static String pinA0str = "A0 (36): ";
static String pinA1str = "A1 (39): ";
static String pinA2str = "A2 (34): ";
static String pinA3str = "A3 (35): ";
static String pinA4str = "A4 (32): ";
static String pinA5str = "A5 (33): ";

static String pinD2str = "D2 (25): ";
static String pinD3str = "D3 (26): ";
static String pinD4str = "D4 (27): ";
static String pinD5str = "D5 (14): ";
static String pinD6str = "D6 (12): ";
static String pinD7str = "D7 (13): ";
// ------- END DEFINITIONS ----------

// ------- PINS ----------
static int morsePin = 2;
static int i2cSdaPin = 21;
static int i2cSclPin = 22;

static int pinA0 = 36;
static int pinA1 = 39;
static int pinA2 = 34;
static int pinA3 = 35;
static int pinA4 = 32;
static int pinA5 = 33;

static int pinD2 = 25;
static int pinD3 = 26;
static int pinD4 = 27;
static int pinD5 = 14;
static int pinD6 = 12;
static int pinD7 = 13;

// ------- END PINS ----------

// --------- Variables ---------

static bool enableBme = false;
String bmeTemp = String("");
String bmeHum = String("");
String bmePress = String("");
Adafruit_BME280 bme; // I2C

int freeHeap = 0;

// --------- END Variables ---------


void setup()
{
  initSerial();
  initWifi();
  initPinModes();
  initSchedules();
  initServer();
  initTelnet();
  initBme();
  doLogic();
  checkFreeRam();
}

// --------------- LOOP ---------------

void loop()
{
  server.handleClient();
  runner.execute();
  telnet.loop();
  doLogic();
}

// --------------- INITS ---------------

void initSerial()
{
  Serial.begin(115200); 
  while(!Serial){} // Waiting for serial connection
  Serial.println();
}

void initWifi()
{
  Serial.println("WiFi init");
  Serial.print("Wifi: ");
  Serial.println(ssid);
  //Serial.print("WifiPW: ");
  //Serial.println(password);
  
  Serial.println("turn wifi off...");
  WiFi.mode(WIFI_OFF);
  delay(10);
  //WiFi.forceSleepBegin();
  delay(200);
  //WiFi.forceSleepWake();
  WiFi.mode(WIFI_STA);
  delay(250);
  
  WiFi.config(INADDR_NONE, INADDR_NONE, INADDR_NONE);    
  delay(200); 
  //WiFi.mode(WIFI_STA);
  Serial.println("setting hostname");
  WiFi.setHostname(deviceName);
  delay(200); 
  Serial.println("Connecting to WiFi..");
  WiFi.begin(ssid, password);
  delay(200); 

  int iCounter = 0;
  int iMax = 30;
  while (WiFi.status() != WL_CONNECTED && iCounter < iMax)
  {
    digitalWrite(morsePin, LOW);
    delay(500);
    digitalWrite(morsePin, HIGH);
    delay(500);
    Serial.print(".");
    iCounter++;
  }
  
  Serial.println(WiFi.localIP()); 
}

void initPinModes()
{
  Serial.println("PIN MODES init");

  pinMode(morsePin, OUTPUT);
  digitalWrite(morsePin, LOW);

  pinMode(pinA0, INPUT);
  pinMode(pinA1, INPUT);
  pinMode(pinA2, INPUT);
  pinMode(pinA3, INPUT);
  pinMode(pinA4, INPUT);
  pinMode(pinA5, INPUT);

  pinMode(pinD2, INPUT);
  pinMode(pinD3, INPUT);
  pinMode(pinD4, INPUT);
  pinMode(pinD5, INPUT);
  pinMode(pinD6, INPUT);
  pinMode(pinD7, INPUT);
}

void initSchedules()
{
  Serial.println("SCHEDULES init");
  
  runner.init();

  runner.addTask(scheduleCheckFreeRam);
  scheduleCheckFreeRam.enable();

  runner.addTask(scheduleWifiReconnectCheck);
  scheduleWifiReconnectCheck.enable();

  runner.addTask(scheduleGetBme);
  scheduleGetBme.enable();

  runner.addTask(scheduleGetValues);
  scheduleGetValues.enable();
}

void initServer()
{
  server.on("/", handleConnect);
  server.onNotFound(handleConnect);
  server.begin();
  Serial.println("HTTP server started");
}

void initTelnet()
{
  // passing on functions for various telnet events
  telnet.onConnect(onTelnetConnect);
  telnet.onConnectionAttempt(onTelnetConnectionAttempt);
  telnet.onReconnect(onTelnetReconnect);
  telnet.onDisconnect(onTelnetDisconnect);
  
  // passing a lambda function
  telnet.onInputReceived([](String str) 
  {
    // checks for a certain command
    if (str == "ping") 
    {
      telnet.println("> pong");
      Serial.println("- Telnet: pong");
    }
  });

  Serial.print("- Telnet: ");
  if (telnet.begin()) 
  {
    Serial.println("running");
  }
  else 
  {
    Serial.println("error.");
    //errorMsg("Will reboot...");
  }
}

void initBme()
{
  // BME 280
  if (enableBme == true)
  {  
    for(int i = 0; i < 10; i++)
    {
      Serial.println("detecting BME280...");
      delay(500);
      
      if (bme.begin(0x76))
      {
        i = 10;
      }
    }
  }
}

// --------------- LOGIC ---------------

void doLogic()
{
}

void getValues()
{
  String lineString = "";

  lineString += "A0:" + String(analogRead(pinA0));
  lineString += ";";

  lineString += "A1:" + String(analogRead(pinA1));
  lineString += ";";

  lineString += "A2:" + String(analogRead(pinA2));
  lineString += ";";

  lineString += "A3:" + String(analogRead(pinA3));
  lineString += ";";

  lineString += "A4:" + String(analogRead(pinA4));
  lineString += ";";

  lineString += "A5:" + String(analogRead(pinA5));
  lineString += ";";


  lineString += "D2:" + String(digitalRead(pinD2));
  lineString += ";";

  lineString += "D3:" + String(digitalRead(pinD3));
  lineString += ";";

  lineString += "D4:" + String(digitalRead(pinD4));
  lineString += ";";

  lineString += "D5:" + String(digitalRead(pinD5));
  lineString += ";";

  lineString += "D6:" + String(digitalRead(pinD6));
  lineString += ";";

  lineString += "D7:" + String(digitalRead(pinD7));
  lineString += ";";
  
  lineString += "#EOL#";

  telnet.println(lineString);
  Serial.println(lineString);
}

// (optional) callback functions for telnet events
void onTelnetConnect(String ip)
{
  Serial.print("- Telnet: ");
  Serial.print(ip);
  Serial.println(" connected");
  telnet.println("\nWelcome " + telnet.getIP());
  telnet.println("(Use ^] + q  to disconnect.)");
}

void onTelnetDisconnect(String ip)
{
  Serial.print("- Telnet: ");
  Serial.print(ip);
  Serial.println(" disconnected");
}

void onTelnetReconnect(String ip)
{
  Serial.print("- Telnet: ");
  Serial.print(ip);
  Serial.println(" reconnected");
}

void onTelnetConnectionAttempt(String ip)
{
  Serial.print("- Telnet: ");
  Serial.print(ip);
  Serial.println(" tried to connected");
}


// --------------- SCHEDULER ---------------

void checkFreeRam()
{
  freeHeap = ESP.getFreeHeap();

  if (ESP.getFreeHeap() < 60000)
  {
    ESP.restart();
  }

  //Serial.println("Free RAM: " + String(ESP.getFreeHeap()));
}

void wifiReconnectCheck()
{
  unsigned long currentMillis = millis();
  // if WiFi is down, try reconnecting every CHECK_WIFI_TIME seconds
  if ((WiFi.status() != WL_CONNECTED) && (currentMillis - previousMillis >=interval))
  {
    Serial.print(millis());
    Serial.println("Reconnecting to WiFi...");
    WiFi.disconnect();
    WiFi.reconnect();
    previousMillis = currentMillis;
  }
}

void getBme()
{
  if (enableBme == true)
  {
    Serial.println("Refresh BME Vals");

    for (int i=0; i < 10; i++)
    {
      double temp = bme.readTemperature();

      if (temp > -40 && temp < 100)
      {
        i = 10;
        bmeTemp = String(temp);
        bmeHum = String(bme.readHumidity());
        bmePress = String(bme.readPressure() / 100.0F);
      }
    }
    
    Serial.println("END - Refresh BME Vals");  
  }
}

// --------------- SERVER ---------------

void handleConnect()
{
  Serial.println("Connect");
  server.send(200, "text/html", SendHTML("")); 
}

String SendHTML(String context)
{
  String ptr = "<!DOCTYPE html> <html>\n";
  ptr +="<head><meta name=\"viewport\" content=\"width=device-width, initial-scale=1.0, user-scalable=no\">\n";
  ptr +="<title>ESP32 DataLogger</title>\n";
  ptr +="<style>html { font-family: Arial; display: inline-block; margin: 0px auto; text-align: center;}\n";
  ptr +="body{margin-top: 50px;} h1 {color: #444444;margin: 50px auto 30px;} h3 {color: #444444;margin-bottom: 50px;}\n";
  ptr +=".button {display: block;width: 80px;background-color: #3498db;border: none;color: white;padding: 13px 30px;text-decoration: none;font-size: 25px;margin: 0px auto 35px;cursor: pointer;border-radius: 4px;}\n";
  ptr +="a, a:active, { color: " + linkColorNormal + "; text-decoration: underline; }\n";
  ptr +="a:visited { color: " + linkColorVisited + "; text-decoration: underline; }\n";
  ptr +="p {font-size: 14px;color: #888;margin-bottom: 10px;}\n";
  ptr +="</style>\n";
  ptr +="</head>\n";
  ptr +="<body>\n";
  ptr +="<h1>ESP32 DataLogger</h1>\n";

  String lineBreak = "<br><br>";


  ptr += lineBreak;

  ptr +="<b>BME 280</b><br>\n";
  ptr += bme280State + ": ";
  if (enableBme == true)
  {
    ptr += bme280Enabled;
    ptr += "<br>";

    ptr += temperatureName + ": ";
    ptr += bmeTemp + temperatureName2 + "<br>";

    ptr += humidityName + ": ";
    ptr += bmeHum + humidityName2 + "<br>";

    ptr += pressureName + ": ";
    ptr += bmePress + pressureName2 + "<br>";
  }
  else
  {
    ptr += bme280Disabled;
    ptr += "<br>";
  }

  ptr += lineBreak;

  ptr += pinA0str + String(analogRead(pinA0)) + "<br>";
  ptr += pinA1str + String(analogRead(pinA1)) + "<br>";
  ptr += pinA2str + String(analogRead(pinA2)) + "<br>";
  ptr += pinA3str + String(analogRead(pinA3)) + "<br>";
  ptr += pinA4str + String(analogRead(pinA4)) + "<br>";
  ptr += pinA5str + String(analogRead(pinA5)) + "<br>";

  ptr += lineBreak;

  ptr += pinD2str + String(analogRead(pinD2)) + "<br>";
  ptr += pinD3str + String(analogRead(pinD3)) + "<br>";
  ptr += pinD4str + String(analogRead(pinD4)) + "<br>";
  ptr += pinD5str + String(analogRead(pinD5)) + "<br>";
  ptr += pinD6str + String(analogRead(pinD6)) + "<br>";
  ptr += pinD7str + String(analogRead(pinD7)) + "<br>";

  ptr += lineBreak;

  ptr += "<b>FreeHeap</b><br>\n";
  ptr += (String)freeHeap + " bytes";

  ptr +="</body>\n";
  ptr +="</html>\n";
  return ptr;
}


